
import bz2
from collections import ChainMap
from datetime import datetime as dt, timedelta
import re
import typing as T

from yarl import URL

from .models import CacheEntry, Pkg, PkgFolder, PkgOptions


EXPIRY = 3600


class BaseResolver:

    __members__ = {}

    def __init__(self, *, manager: 'Manager'):
        self.manager = manager
        self.synced = False

    def __init_subclass__(cls, origin: str):
        cls.__members__[origin] = cls
        cls.origin = origin

        orig_sync = cls.sync
        async def _sync(self) -> None:
            if not self.synced:
                await orig_sync(self)
                self.synced = True
        cls.sync = _sync

    @classmethod
    def decompose_url(cls, url: str) -> T.Optional[T.Tuple[str, str]]:
        """Break a URL down into its component `origin` and `id`."""

    async def sync(self) -> None:
        """Serves as a deferred, asynchronous `__init__`.  Do any
        preprocessing here if necessary, including writing to the cache.
        """
    async def resolve(self, id_or_slug: str, *,
                      strategy: str) -> Pkg:
        """Turn an ID or slug into a `models.Pkg`."""
        raise NotImplementedError


class _CurseResolver(BaseResolver,
                     origin='curse'):

    _data: dict

    _json_dump_url = 'https://clientupdate-v6.cursecdn.com/feed/addons/1/v10/'\
                     '{freq}.json.bz2'
    _json_date_url = 'https://clientupdate-v6.cursecdn.com/feed/addons/1/v10/'\
                     '{freq}.json.bz2.txt'

    _freqs = {'hourly', 'daily', 'weekly', 'complete'}

    _re_curse_url = re.compile(r'(?P<id>\d+)-(?P<slug>[a-z_-]+)')

    @classmethod
    def _slug_from_url(cls, url: str) -> str:
        name = URL(url).name
        match = cls._re_curse_url.match(name)
        if match:
            return match.group('slug')
        return name

    @classmethod
    def decompose_url(cls, url: str) -> T.Optional[T.Tuple[str, str]]:
        url = URL(url)
        if url.host in {'wow.curseforge.com', 'www.wowace.com'} \
                and len(url.parts) > 2 \
                and url.parts[1] == 'projects':
            return (cls.origin, url.parts[2])
        elif url.host == 'mods.curse.com':
            if len(url.parts) > 3 \
                    and url.parts[1:3] == ('addons', 'wow'):
                slug = url.parts[3]
                match = cls._re_curse_url.match(slug)
                if match:
                    slug = match.group('id')
                return (cls.origin, slug)
            elif len(url.parts) == 3 \
                    and url.parts[1] == 'project':
                return (cls.origin, url.parts[2])

    async def sync(self) -> None:
        async def _sync(freq):
            entry = self.manager.db.query(CacheEntry).get((self.origin, freq))
            if not entry or ((entry.date_retrieved + timedelta(seconds=EXPIRY)) <
                             dt.now()):
                async with self.manager.client.get(self._json_date_url
                                                   .format(freq=freq)) as resp:
                    dt_ = dt.fromtimestamp(int(await resp.text()) / 1000)
                if not entry or entry.date_updated != dt_:
                    async with self.manager.client.get(self._json_dump_url
                                                       .format(freq=freq)) as resp:
                        payload = await resp.read()
                    payload = await self.manager.block_in_thread(
                        lambda: bz2.decompress(payload).decode())
                    new_entry = CacheEntry(origin=self.origin,
                                           id=freq,
                                           date_retrieved=dt.now(),
                                           date_updated=dt_,
                                           contents=payload)
                    entry = new_entry.replace(self.manager.db, entry)
            return entry

        data = await self.manager.gather((_sync(f) for f in self._freqs),
                                         show_progress=self.manager.show_progress,
                                         label='Updating Curse cache')
        data = sorted(data, key=lambda e: e.date_updated, reverse=True)
        data = ChainMap(*({str(e['Id']): e
                           for e in c.contents['data'] if e['PackageType'] == 1}
                          for c in data),
                        *({self._slug_from_url(e['WebSiteURL']): e
                           for e in c.contents['data'] if e['PackageType'] == 1}
                          for c in data))
        self._data = data

    async def resolve(self,
                      id_or_slug: str,
                      *,
                      strategy: str) -> Pkg:
        try:
            proj = self._data[id_or_slug]
        except KeyError:
            # This really shouldn't happen but projects are sometimes missing
            # from the data dumps - maybe there's another dump we're yet to
            # uncover?
            raise self.manager.PkgNonexistent

        if strategy == 'canonical':
            file_id = proj['DefaultFileId']
        elif strategy == 'latest':
            file_id = max(f['Id'] for f in proj['LatestFiles']
                          if f['IsAlternate'] is False)
        file = next(f for f in proj['LatestFiles'] if f['Id'] == file_id)

        return Pkg(origin=self.origin,
                   id=proj['Id'],
                   slug=self._slug_from_url(proj['WebSiteURL']),
                   name=proj['Name'],
                   description=proj['Summary'],
                   url=proj['WebSiteURL'],
                   file_id=file['Id'],
                   download_url=file['DownloadURL'],
                   date_published=file['FileDate'],
                   folders=[PkgFolder(path=self.manager.config.addon_dir/m['Foldername'])
                            for m in file['Modules']],
                   version=file['FileName'],
                   options=PkgOptions(strategy=strategy))


class _WowiResolver(BaseResolver,
                    origin='wowi'):

    _data: dict

    _json_dump_url = 'https://api.mmoui.com/v3/game/WOW/filelist.json'
    _details_api_endpoint = 'https://api.mmoui.com/v3/game/WOW/filedetails/'

    _re_head = re.compile(r'^info')
    _re_tail = re.compile(r'\.html$')

    _re_addon_url = re.compile(r'(?:download|info)(\d+)')

    @classmethod
    def _slugify(cls, url: str) -> str:
        return cls._re_tail.sub('', cls._re_head.sub('', URL(url).name))

    @classmethod
    def decompose_url(cls, url: str) -> T.Optional[T.Tuple[str, str]]:
        url = URL(url)
        if url.host in {'wowinterface.com', 'www.wowinterface.com'} \
                and len(url.parts) == 3 \
                and url.parts[1] == 'downloads':
            match = cls._re_addon_url.match(url.name)
            if match:
                return (cls.origin, match.group(1))

    async def sync(self) -> None:
        entry = self.manager.db.query(CacheEntry).get((self.origin, self.origin))
        if not entry or ((entry.date_retrieved + timedelta(seconds=EXPIRY)) <
                         dt.now()):
            async with self.manager.client.get(self._json_dump_url) as resp:
                data = await resp.read()
            new_entry = CacheEntry(origin=self.origin, id=self.origin,
                                   date_retrieved=dt.now(), contents=data)
            entry = new_entry.replace(self.manager.db, entry)

        self._data = {e['UID']: e for e in entry.contents}

    async def resolve(self,
                      id_or_slug: str,
                      *,
                      strategy: str) -> Pkg:
        file_id = id_or_slug if id_or_slug.isnumeric() else \
            id_or_slug.partition('-')[0]
        try:
            file = self._data[file_id]
        except KeyError:
            raise self.manager.PkgNonexistent

        async with self.manager.client.get(f'{self._details_api_endpoint}'
                                           f'/{file_id}.json') as resp:
            details, = await resp.json()
        if file['UIDate'] != details['UIDate']:
            raise self.manager.CacheObsolete

        return Pkg(origin=self.origin,
                   id=file['UID'],
                   slug=self._slugify(file['UIFileInfoURL']),
                   name=file['UIName'],
                   description=details['UIDescription'],
                   url=file['UIFileInfoURL'],
                   file_id=details['UIMD5'],
                   download_url=details['UIDownload'],
                   date_published=details['UIDate'],
                   folders=[PkgFolder(path=self.manager.config.addon_dir/f)
                            for f in file['UIDir']],
                   version=details['UIVersion'],
                   options=PkgOptions(strategy=strategy))




from collections import namedtuple
import io
import os
from pathlib import Path
import typing as T
import zipfile

from click._termui_impl import ProgressBar as _ProgressBar, \
                               BEFORE_BAR, AFTER_BAR


class ExtractConflict(Exception):

    def __init__(self, conflicting_folders):
        super().__init__()
        self.conflicting_folders = conflicting_folders


class Archive:

    ExtractConflict = ExtractConflict

    def __init__(self, payload: bytes):
        self._archive = zipfile.ZipFile(io.BytesIO(payload))

    def extract(self, parent_folder: Path,
                *,
                overwrite: T.Union[bool, T.Set[str]]=False) -> None:
        if overwrite is not True:
            conflicts = ({Path(p).parts[0] for p in self._archive.namelist()} &
                         ({f.name for f in parent_folder.iterdir()} -
                          (overwrite or set())))
            if conflicts:
                raise ExtractConflict(conflicts)
        self._archive.extractall(parent_folder)


_TocEntry = namedtuple('_TocEntry', 'key value')

class TocReader:
    """Extracts key–value pairs from TOC files."""

    def __init__(self, toc_file_path: Path):
        entries = (e.lstrip('# ').partition(': ')
                   for e in toc_file_path.read_text().splitlines()
                   if e.startswith('## '))
        entries = {e[0]: e[2] for e in entries}
        self.entries = entries

    def __getitem__(self, keys: T.Union[str, T.Tuple[str]]) -> _TocEntry:
        if isinstance(keys, tuple):
            try:
                return next(filter(lambda i: i.value,
                                   (self.__getitem__(k) for k in keys)))
            except StopIteration:
                keys = keys[0]
        return _TocEntry(keys, self.entries.get(keys))


class ProgressBar(_ProgressBar):
    """A `ProgressBar` subclass that clears its output upon completion.

    Only use this if you expect an operation to take a _perceptibly_
    long time (e.g. network transfers) -- not every single thing we do
    merits a progress bar.
    """

    def __init__(self, **kwargs):
        super().__init__(**{'iterable': None,
                            'bar_template': '%(label)s  [%(bar)s]  %(info)s',
                            **kwargs})

    def render_finish(self) -> None:
        if self.is_hidden:
            return

        self.file.write(BEFORE_BAR + (' ' * self.max_width) + AFTER_BAR +
                        ('' if os.name == 'nt' else '\x1b[A'))
        self.file.flush()




from datetime import datetime
import json
from pathlib import Path

import pydantic
from sqlalchemy import Column, String, DateTime, Enum, ForeignKeyConstraint, \
                       TypeDecorator, or_
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship


def _declarative_constructor(self, **kwargs):
    for k, v in _BaseCoercer.__members__[self.__class__]\
                            .parse_obj(kwargs):
        setattr(self, k, v)


ModelBase = declarative_base(constructor=_declarative_constructor)


class _MetaBase(pydantic.BaseModel.__class__):

    __members__ = {}

    def __new__(mcs, name, bases, namespace, **kwargs):
        cls = super().__new__(mcs, name, bases, namespace)
        if 'coerces' in kwargs:
            mcs.__members__[kwargs['coerces']] = cls
        return cls


class _BaseCoercer(pydantic.BaseModel, metaclass=_MetaBase):
    """The coercer is used inside the declarative constructor to type-cast
    values _prior to_ inserts.  SQLAlchemy delegates this function to
    the DB API - see https://stackoverflow.com/a/8980982.  We need this to
    happen in advance to be able to compare values with their in-database
    equivalents without having to wrap every single one of them inside
    `str()` (yuck).  It also saves us the trouble of having to manually
    parse dates and the like.
    """

    class Config:
        allow_extra = True
        max_anystr_length = 2 ** 32


class _PathType(TypeDecorator):

    impl = String

    def process_bind_param(self, value, dialect):
        if value is not None:
            return str(value)

    def process_result_value(self, value, dialect):
        if value is not None:
            return Path(value)


class _JsonType(TypeDecorator):

    impl = String

    def process_result_value(self, value, dialect):
        if value is not None:
            return json.loads(value)


class _ConvenienceMethodsMixin:

    def insert(self, session):
        session.add(self)
        session.commit()
        return self

    def replace(self, session, other=None):
        if other:
            session.delete(other)
            session.commit()
        return self.insert(session)

    def delete(self, session):
        session.delete(self)
        session.commit()


class Pkg(ModelBase,
          _ConvenienceMethodsMixin):

    __tablename__ = 'pkg'

    origin = Column(String, primary_key=True)
    id = Column(String, primary_key=True)
    slug = Column(String, nullable=False)
    name = Column(String, nullable=False)
    description = Column(String, nullable=False)
    url = Column(String, nullable=False)
    file_id = Column(String, nullable=False)
    download_url = Column(String, nullable=False)
    date_published = Column(DateTime, nullable=False)
    version = Column(String, nullable=False)
    folders = relationship('PkgFolder', cascade='all, delete-orphan',
                           backref='pkg')
    options = relationship('PkgOptions', cascade='all, delete-orphan',
                           uselist=False)

    @classmethod
    def unique(cls, origin, id_or_slug, session):
        return session.query(cls).filter(cls.origin == origin,
                                         or_(cls.id == id_or_slug, cls.slug == id_or_slug))\
                      .first()


class _PkgCoercer(_BaseCoercer, coerces=Pkg):

    origin: str
    id: str
    slug: str
    name: str
    description: str
    url: str
    file_id: str
    download_url: str
    date_published: datetime
    version: str


class PkgFolder(ModelBase):

    __tablename__ = 'pkg_folder'
    __table_args__ = (ForeignKeyConstraint(['pkg_origin', 'pkg_id'],
                                           ['pkg.origin', 'pkg.id']),)

    path = Column(_PathType, primary_key=True)  # TODO: migrate paths to basenames
    pkg_origin = Column(String, nullable=False)
    pkg_id = Column(String, nullable=False)


class _PkgFolderCoercer(_BaseCoercer, coerces=PkgFolder):

    path: Path


class PkgOptions(ModelBase):

    __tablename__ = 'pkg_options'
    __table_args__ = (ForeignKeyConstraint(['pkg_origin', 'pkg_id'],
                                           ['pkg.origin', 'pkg.id']),)

    strategy = Column(Enum('canonical', 'latest'), nullable=False)
    pkg_origin = Column(String, primary_key=True)
    pkg_id = Column(String, primary_key=True)


class _PkgOptionsCoercer(_BaseCoercer, coerces=PkgOptions):

    strategy: str


class CacheEntry(ModelBase,
                 _ConvenienceMethodsMixin):

    __tablename__ = 'cache'

    origin = Column(String, primary_key=True)
    id = Column(String, primary_key=True)
    date_updated = Column(DateTime)
    date_retrieved = Column(DateTime, nullable=False)
    contents = Column(_JsonType, nullable=False)


class _CacheEntryCoercer(_BaseCoercer, coerces=CacheEntry):

    origin: str
    id: str
    date_updated: datetime = None
    date_retrieved: datetime
    contents: str




import asyncio
from concurrent.futures import ThreadPoolExecutor
from contextlib import ExitStack
import shutil
import typing as T

from aiohttp import ClientSession, TCPConnector
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from .config import Config
from .models import ModelBase, Pkg, PkgFolder
from .resolvers import BaseResolver
from .utils import Archive, ProgressBar


_UA_STRING = 'instawow (https://github.com/layday/instawow)'


def _init_loop():
    try:
        import uvloop
    except ImportError:
        return asyncio.get_event_loop()
    else:
        return uvloop.new_event_loop()


def _dedupe(it):
    list_ = list(it)
    return sorted(set(list_), key=list_.index)


async def _intercept_fut(index, fut, return_exceptions):
    try:
        result = await fut
    except Exception as e:
        if return_exceptions:
            result = e
        else:
            raise
    return index, result


def _intercept_result(result):
    if callable(result):
        try:
            result = result()
        except Exception as e:
            result = e
    return result


class ManagerResult(Exception):
    pass


class PkgAlreadyInstalled(ManagerResult):
    pass


class PkgConflictsWithInstalled(ManagerResult):

    def __init__(self, pkg):
        super().__init__()
        self.conflicting_pkg = pkg


class PkgConflictsWithPreexisting(ManagerResult):
    pass


class PkgNonexistent(ManagerResult):
    pass


class PkgNotInstalled(ManagerResult):
    pass


class PkgOriginInvalid(ManagerResult):
    pass


class PkgUpToDate(ManagerResult):
    pass


class PkgInstalled(ManagerResult):

    def __init__(self, pkg):
        super().__init__()
        self.new_pkg = pkg


class PkgUpdated(ManagerResult):

    def __init__(self, pkgs):
        super().__init__()
        self.old_pkg, self.new_pkg = pkgs


class PkgModified(ManagerResult):

    def __init__(self, key, value):
        super().__init__()
        self.key = key
        self.value = value


class PkgRemoved(ManagerResult):
    pass


class CacheObsolete(ManagerResult):
    pass


class _Runner:

    def __init__(self, manager):
        self.manager = manager

    def __getattr__(self, name):
        return lambda *a, **kw: (self.manager._loop.run_until_complete
                                 (getattr(self.manager, name)(*a, **kw)))


class Manager:

    ManagerResult               = ManagerResult
    PkgInstalled                = PkgInstalled
    PkgUpdated                  = PkgUpdated
    PkgModified                 = PkgModified
    PkgRemoved                  = PkgRemoved
    PkgAlreadyInstalled         = PkgAlreadyInstalled
    PkgConflictsWithPreexisting = PkgConflictsWithPreexisting
    PkgConflictsWithInstalled   = PkgConflictsWithInstalled
    PkgNonexistent              = PkgNonexistent
    PkgNotInstalled             = PkgNotInstalled
    PkgOriginInvalid            = PkgOriginInvalid
    PkgUpToDate                 = PkgUpToDate
    CacheObsolete               = CacheObsolete

    def __init__(self,
                 *,
                 config: Config,
                 loop: asyncio.BaseEventLoop=_init_loop(),
                 show_progress: bool=True):
        self.config = config
        self.show_progress = show_progress

        db_engine = create_engine(f'sqlite:///{config.config_dir/config.db_name}')
        ModelBase.metadata.create_all(db_engine)
        self.db = sessionmaker(bind=db_engine)()
        self.client = ClientSession(connector=TCPConnector(limit_per_host=10, loop=loop),
                                    headers={'User-Agent': _UA_STRING}, loop=loop)
        self.resolvers = {n: r(manager=self)
                          for n, r in BaseResolver.__members__.items()}
        self.runner = _Runner(self)

        self._loop = loop
        self._resolver_lock = asyncio.Lock(loop=loop)
        self._tpes = [ThreadPoolExecutor(max_workers=1), ThreadPoolExecutor(max_workers=1)]

    async def block_in_thread(self, fn: T.Callable, *, channel: int=0) -> T.Any:
        """Execute a function in a separate thread.  Successive calls to this
        method are queued by virtue of reusing the same thread.
        """
        return await self._loop.run_in_executor(self._tpes[channel], fn)

    async def gather(self, it, *,
                     return_exceptions: bool=False, show_progress: bool=False,
                     **kwargs) -> list:
        """Execute coroutines concurrently and gather their results.
        This displays a progress bar in the command line when
        `show_progress=True`.
        """
        if not show_progress:
            return await asyncio.gather(*it, loop=self._loop,
                                        return_exceptions=return_exceptions)

        futures = [_intercept_fut(i, f, return_exceptions)
                   for i, f in enumerate(it)]
        results = [None] * len(futures)
        with ProgressBar(length=len(futures), **kwargs) as bar:
            for result in asyncio.as_completed(futures, loop=self._loop):
                results.__setitem__(*(await result))
                bar.update(1)
        return results

    def __enter__(self) -> 'Manager':
        return self

    def __exit__(self, *_e):
        self.close()

    def close(self) -> None:
        self.client.close()

    async def _resolve(self, origin: str, id_or_slug: str,
                       strategy: str) -> T.Union[ManagerResult, Pkg]:
        if origin not in self.resolvers:
            raise PkgOriginInvalid

        async with self._resolver_lock:
            await self.resolvers[origin].sync()
        return await self.resolvers[origin].resolve(id_or_slug,
                                                    strategy=strategy)

    async def _prepare_install(self, origin: str, id_or_slug: str,
                               strategy: str, overwrite: bool) -> T.Callable:
        if Pkg.unique(origin, id_or_slug, self.db):
            raise PkgAlreadyInstalled

        new_pkg = await self._resolve(origin, id_or_slug, strategy)
        async with self.client.get(new_pkg.download_url) as resp:
            payload = await resp.read()

        def _finalise():
            folder_conflict = \
                self.db.query(PkgFolder)\
                       .filter(PkgFolder.path.in_([f.path for f in new_pkg.folders]))\
                       .first()
            if folder_conflict:
                raise PkgConflictsWithInstalled(folder_conflict.pkg)
            try:
                Archive(payload).extract(self.config.addon_dir,
                                         overwrite=overwrite)
            except Archive.ExtractConflict:
                raise PkgConflictsWithPreexisting
            else:
                return PkgInstalled(new_pkg.insert(self.db))
        return _finalise

    async def _prepare_update(self, origin: str, id_or_slug: str) -> T.Callable:
        old_pkg = Pkg.unique(origin, id_or_slug, self.db)
        if not old_pkg:
            raise PkgNotInstalled

        new_pkg = await self._resolve(origin, id_or_slug, old_pkg.options.strategy)
        if old_pkg.file_id == new_pkg.file_id:
            raise PkgUpToDate
        async with self.client.get(new_pkg.download_url) as resp:
            payload = await resp.read()

        def _finalise():
            folders = [f.path for f in new_pkg.folders]
            folder_conflict = \
                self.db.query(PkgFolder)\
                       .filter(PkgFolder.path.in_(folders),
                               PkgFolder.pkg_origin != new_pkg.origin,
                               PkgFolder.pkg_id != new_pkg.id)\
                       .first()
            if folder_conflict:
                raise PkgConflictsWithInstalled(folder_conflict.pkg)

            try:
                Archive(payload).extract(self.config.addon_dir,
                                         overwrite={f.name for f in folders})
            except Archive.ExtractConflict:
                raise PkgConflictsWithPreexisting
            else:
                return PkgUpdated((old_pkg, new_pkg.replace(self.db, old_pkg)))
        return _finalise

    def resolve_many(self, values: T.Iterable[tuple]) -> T.List[T.Union[ManagerResult, Pkg]]:
        return self.runner.gather((self._resolve(*t) for t in _dedupe(values)),
                                  return_exceptions=True,
                                  show_progress=self.show_progress,
                                  label='Resolving')

    def install_many(self, values: T.Iterable[tuple]) -> T.List[ManagerResult]:
        results = self.runner.gather((self._prepare_install(*q)
                                      for q in _dedupe(values)),
                                     return_exceptions=True,
                                     show_progress=self.show_progress,
                                     label='Resolving add-ons')
        with ExitStack() as stack:
            if self.show_progress:
                results = stack.enter_context(ProgressBar(iterable=results,
                                                          label='Installing'))
            return [_intercept_result(r) for r in results]

    def update_many(self, values: T.Iterable[tuple]) -> T.List[ManagerResult]:
        results = self.runner.gather((self._prepare_update(*p)
                                      for p in _dedupe(values)),
                                     return_exceptions=True,
                                     show_progress=self.show_progress,
                                     label='Resolving add-ons')
        with ExitStack() as stack:
            if self.show_progress:
                results = stack.enter_context(ProgressBar(iterable=results,
                                                          label='Updating'))
            return [_intercept_result(r) for r in results]

    def remove(self, origin: str, id_or_slug: str) -> PkgRemoved:
        pkg = Pkg.unique(origin, id_or_slug, self.db)
        if not pkg:
            raise PkgNotInstalled

        for folder in pkg.folders:
            shutil.rmtree(folder.path)
        return PkgRemoved(pkg.delete(self.db))
